clc;
clear;

C = 3000;
n = 100;
T = 50;

r = 2;
h = 4;
J = 10;
rep = 100;                        %% # of bootstrap replications

a = [1, 1, 1]';

phi = 0.7;

p = 2 ;                           %% Parzen exponential characteristic
K_p = -6 ;                        %% Exponential characteristic of Parzen Kernel            

rho = [0.3, 0.3^2]';

O = min(sqrt(n),sqrt(T));

dn = [0.5:0.5:20]'; 
dn_true = [0.1, 1:39]'; 

cor = zeros(n,1);

for j = 1 : J;
    cor(j,1) = phi^(j-1);
end;

Omega_sq = toeplitz(cor);
Omega = chol(Omega_sq);

dn_star = zeros(C,1);
n_sub = round(sqrt(min(n,T)));

D1_true = repmat([1:n]',1,n);
D2_true = repmat([1:n],n,1);
D_true = abs(D1_true-D2_true);

ln_temp_true = zeros(size(dn_true,1),1);                                %% average number of individual withdin bandwidth

for s = 1:size(dn_true,1) ;
    for ss = 1:n;
        ln_temp_true(ss,s) = sum(D_true(:,ss)<=dn_true(s,1))-1;
    end;
end;

ln_true = mean(ln_temp_true)';

dn_star_true = zeros(C,1);

cnt = zeros(C,1);                    % proposed method + data driven distance
cnt_cm = zeros(C,1);                 % _cm: conditional mean prediction interval
cnt_true = zeros(C,1);               % proposed method + true distance
cnt_cm_true = zeros(C,1);
cnt_con = zeros(C,1);                % HetoR method
cnt_con_cm = zeros(C,1);

cnt_HAC_no_order = zeros(C,1);       % Bai & Ng's CS-HAC without ordering
cnt_HAC_cm_no_order = zeros(C,1);
cnt_HAC_yes_order = zeros(C,1);      % Bai & Ng's CS-HAC with ordering
cnt_HAC_cm_yes_order = zeros(C,1);

for c = 1 : C;

    sig_v_t = rand(1,n)+0.5;
    sig_v = repmat(sig_v_t,T,1);
    sig_u_t = rand(1,r)+0.5;
    sig_u = repmat(sig_u_t,T,1);
    u = sqrt(sig_u).*randn(T,r);

    f_temp = zeros(T+1,r);
    for t = 1 : T
        f_temp(t+1,:) = rho'.*f_temp(t,:) + sqrt(1-rho'.^2).*u(t,:);
    end;
    f = f_temp(2:end,:);

    lambda = rand(n,r);

    v = sqrt(sig_v).*randn(T,n);
    e = v*Omega;

    X = f*lambda' + e; 

    error = randn(T,1);
    yh_cm = [ones(T,1) f]*a; 
    yh = yh_cm + error; 

    %% Estimation and prediction

    [evec_all_temp,eval_all_temp] = eig(X*X'/(T*n));
    [eval_all,ind] = sort(diag(eval_all_temp),'descend');

    evec_all = evec_all_temp(:,ind');
    f_til_temp = evec_all(:,1:r);
    diag_f_til_temp = diag(f_til_temp'*f_til_temp/T);
    f_til = f_til_temp./sqrt(repmat(diag_f_til_temp',T,1));
    lam_til = X'*f_til/T;
    V_til = diag(eval_all(1:r,1));

    z_hat = [ones(T,1), f_til];
    b_hat = z_hat(1:T-h,:)\yh(1:T-h);
    a_hat = b_hat(2:end,1);
    yh_cm_hat = z_hat*b_hat;                % Conditional Mean prediction

    eps_hat = yh(1:T-h) - yh_cm_hat(1:T-h); 
    varhat_eps = eps_hat'*eps_hat/T;

    zeps_hat = repmat(eps_hat,1,size(z_hat,2)).*z_hat(1:T-h,:);

    AV_delta =  ((z_hat(1:T-h,:)'*z_hat(1:T-h,:)/T)\(zeps_hat'*zeps_hat/T))/(z_hat(1:T-h,:)'*z_hat(1:T-h,:)/T);

    e_til = X - f_til*lam_til';
    eT_til = e_til(end,:)';

    eta_T = lam_til.*repmat(eT_til,1,r);
    eta_T1 = eta_T(:,1);
    eta_T2 = eta_T(:,2);

    %% Construct data driven distance
    dm_e_til = e_til;
    cov_hat_temp = dm_e_til'*dm_e_til/T;
    sd_hat = sqrt(diag(cov_hat_temp));

    corr_hat = cov_hat_temp./(sd_hat*sd_hat');
    D_hat = abs(min(1./abs(corr_hat),100)) - ones(n,n) ;

    ln_temp = zeros(size(dn,1),1);                                %% average number of individual withdin bandwidth

    for s = 1:size(dn,1) ;
        for ss = 1:n;
            ln_temp(ss,s) = sum(D_hat(:,ss)<=dn(s,1))-1;
        end;
    end;

    ln_hat = mean(ln_temp)';

    %% Bootstrap based bandwidth selection

    mae_one_n = zeros(rep,size(dn,1));
    mae_one_n_true = zeros(rep,size(dn,1));

    for k0 = 1:rep

        rsp = randsample(1:T,T,true); 
        per = randsample(-1:2:1,T,true)';
        per2 = repmat(per,1,n);

        e_star = e_til(rsp,:).*per2;

        X_star = f_til*lam_til' + e_star; 

        [evec_all_temp,eval_all_temp] = eig(X_star*X_star'/(T*n));
        [eval_all,ind] = sort(diag(eval_all_temp),'descend');

        evec_all = evec_all_temp(:,ind');
        f_til_temp = evec_all(:,1:r);
        diag_f_til_temp = diag(f_til_temp'*f_til_temp/T);
        f_til_star = f_til_temp./sqrt(repmat(diag_f_til_temp',T,1));
        lam_til_star = X_star'*f_til_star/T;
        V_til_star = diag(eval_all(1:r,1));

        e_til_star = X_star - f_til_star*lam_til_star';

        for i = 1:size(dn,1);

            % Use data driven distance

            D_n = D_hat/dn(i,1) ;
            k_n = (1-6*D_n.^2 + 6*abs(D_n).^3).*(abs(D_n) <= 0.5) + 2*(1-abs(D_n)).^3 .*(abs(D_n) > 0.5).*(abs(D_n) <= 1);

            gamma_n11 = 0;
            gamma_n12 = 0;
            gamma_n22 = 0;

            % Use true distance

            D_n_true = D_true/dn_true(i,1) ;
            k_n_true = (1-6*D_n_true.^2 + 6*abs(D_n_true).^3).*(abs(D_n_true) <= 0.5) + 2*(1-abs(D_n_true)).^3 .*(abs(D_n_true) > 0.5).*(abs(D_n_true) <= 1);

            gamma_n_true11 = 0;
            gamma_n_true12 = 0;
            gamma_n_true22 = 0;

                for t = 1:T;

                    et_til_star = e_til_star(t,:)';
                    eta_t_star = lam_til_star.*repmat(et_til_star,1,r);
                    eta_t1_star = eta_t_star(:,1);
                    eta_t2_star = eta_t_star(:,2);

                    gamma_n11 = gamma_n11 + sum(sum((eta_t1_star*eta_t1_star').*k_n))/n;
                    gamma_n12 = gamma_n12 + sum(sum((eta_t1_star*eta_t2_star').*k_n))/n;
                    gamma_n22 = gamma_n22 + sum(sum((eta_t2_star*eta_t2_star').*k_n))/n;

                    gamma_n_true11 = gamma_n_true11 + sum(sum((eta_t1_star*eta_t1_star').*k_n_true))/n;
                    gamma_n_true12 = gamma_n_true12 + sum(sum((eta_t1_star*eta_t2_star').*k_n_true))/n;
                    gamma_n_true22 = gamma_n_true22 + sum(sum((eta_t2_star*eta_t2_star').*k_n_true))/n;

                end;

            gamma_n = [gamma_n11, gamma_n12
                gamma_n12, gamma_n22]/T;

            AV_f_star = (V_til\gamma_n)/V_til;    
            mae_one_n(k0,i) = -a_hat'*AV_f_star*a_hat; 

            gamma_n_true = [gamma_n_true11, gamma_n_true12
                gamma_n_true12, gamma_n_true22]/T;

            AV_f_true_star = (V_til\gamma_n_true)/V_til;    
            mae_one_n_true(k0,i) = -a_hat'*AV_f_true_star*a_hat; 

        end;

    end;

    mae_hat = mean(mae_one_n)' - min(mean(mae_one_n)) *(ln_hat > min(n,T)^(2/3));
    mae_true = mean(mae_one_n_true)' - min(mean(mae_one_n_true)) *(ln_true > min(n,T)^(2/3));

    % Use data driven distance

    x1 = find(mae_hat(:,1)==min(mae_hat(:,1)));
    dn_star(c,1) = dn(min(x1),1) ;
    D_star = D_hat/dn_star(c,1) ;

    k = (1-6*D_star.^2 + 6*abs(D_star).^3).*(abs(D_star) <= 0.5) + 2*(1-abs(D_star)).^3 .*(abs(D_star) > 0.5).*(abs(D_star) <= 1);

    gamma_hat11 = 0;
    gamma_hat12 = 0;
    gamma_hat22 = 0;

    for t = 1:T;
        et_til = e_til(t,:)';
        eta_t = lam_til.*repmat(et_til,1,r);
        eta_t1 = eta_t(:,1);
        eta_t2 = eta_t(:,2);

        gamma_hat11 = gamma_hat11 + sum(sum((eta_t1*eta_t1').*k))/n;
        gamma_hat12 = gamma_hat12 + sum(sum((eta_t1*eta_t2').*k))/n;
        gamma_hat22 = gamma_hat22 + sum(sum((eta_t2*eta_t2').*k))/n;
    end;

    gamma_hat = [gamma_hat11, gamma_hat12
        gamma_hat12, gamma_hat22]/T;

    [eve1,eva1] = eig(gamma_hat);
    eva_plus1 = diag(max(0.000001,diag(eva1)));
    gamma_hat = eve1*eva_plus1*eve1';

    AV_f = (V_til\gamma_hat)/V_til;    
    B2 = z_hat(end,:)*AV_delta*z_hat(end,:)'/T + a_hat'*AV_f*a_hat/n;

    PI_cm = [yh_cm_hat(T,1) - 1.96*sqrt(B2), yh_cm_hat(T,1) + 1.96*sqrt(B2)];
    PI = [yh_cm_hat(T,1) - 1.96*sqrt(B2 + varhat_eps), yh_cm_hat(T,1) + 1.96*sqrt(B2 + varhat_eps)];

    cnt_cm(c,1) = (yh_cm(T,1)>PI_cm(1,1))*(yh_cm(T,1)<PI_cm(1,2));     % Count rejections
    cnt(c,1) = (yh(T,1)>PI(1,1))*(yh(T,1)<PI(1,2));

    % Use true distance

    x1_true = find(mae_true(:,1)==min(mae_true(:,1)));
    dn_star_true(c,1) = dn_true(min(x1_true),1) ;
    D_star_true = D_true/dn_star_true(c,1) ;

    k_true = (1-6*D_star_true.^2 + 6*abs(D_star_true).^3).*(abs(D_star_true) <= 0.5) + 2*(1-abs(D_star_true)).^3 .*(abs(D_star_true) > 0.5).*(abs(D_star_true) <= 1);

    gamma_true11 = 0;
    gamma_true12 = 0;
    gamma_true22 = 0;

    for t = 1:T;
        et_til = e_til(t,:)';
        eta_t = lam_til.*repmat(et_til,1,r);
        eta_t1 = eta_t(:,1);
        eta_t2 = eta_t(:,2);

        gamma_true11 = gamma_true11 + sum(sum((eta_t1*eta_t1').*k_true))/n;
        gamma_true12 = gamma_true12 + sum(sum((eta_t1*eta_t2').*k_true))/n;
        gamma_true22 = gamma_true22 + sum(sum((eta_t2*eta_t2').*k_true))/n;
    end;

    gamma_true = [gamma_true11, gamma_true12
        gamma_true12, gamma_true22]/T;

    [eve2,eva2] = eig(gamma_true);
    eva_plus2 = diag(max(0.000001,diag(eva2)));
    gamma_true = eve2*eva_plus2*eve2';

    AV_f_true = (V_til\gamma_true)/V_til;    
    B2_true = z_hat(end,:)*AV_delta*z_hat(end,:)'/T + a_hat'*AV_f_true*a_hat/n;

    PI_cm_true = [yh_cm_hat(T,1) - 1.96*sqrt(B2_true), yh_cm_hat(T,1) + 1.96*sqrt(B2_true)];
    PI_true = [yh_cm_hat(T,1) - 1.96*sqrt(B2_true + varhat_eps), yh_cm_hat(T,1) + 1.96*sqrt(B2_true + varhat_eps)];

    cnt_cm_true(c,1) = (yh_cm(T,1)>PI_cm_true(1,1))*(yh_cm(T,1)<PI_cm_true(1,2));   % Count rejections
    cnt_true(c,1) = (yh(T,1)>PI_true(1,1))*(yh(T,1)<PI_true(1,2));

    %%  Heteroskedasticity robust variance estimator based PIs
    gamma_con = eta_T'*eta_T/n;

    AV_f_con = (V_til\gamma_con)/V_til;    
    B2_con = z_hat(end,:)*AV_delta*z_hat(end,:)'/T + a_hat'*AV_f_con*a_hat/n;

    PI_cm_con = [yh_cm_hat(T,1) - 1.96*sqrt(B2_con), yh_cm_hat(T,1) + 1.96*sqrt(B2_con)];
    PI_con = [yh_cm_hat(T,1) - 1.96*sqrt(B2_con + varhat_eps ), yh_cm_hat(T,1) + 1.96*sqrt(B2_con + varhat_eps)];

    cnt_con_cm(c,1) = (yh_cm(T,1)>PI_cm_con(1,1))*(yh_cm(T,1)<PI_cm_con(1,2));
    cnt_con(c,1) = (yh(T,1)>PI_con(1,1))*(yh(T,1)<PI_con(1,2));

    %%  CS-HAC variance (in Bai and Ng (2006)) based PIs - with and without ordering

    gamma_HAC_temp2_no_order = [0, 0
        0, 0];
    gamma_HAC_temp_no_order = [0, 0
        0, 0];

    gamma_HAC_temp_yes_order = [0,0
        0,0];
    gamma_HAC_temp2_yes_order = [0,0
        0,0];

    for o = 1:O;

        sp = randsample(n,n_sub);
        lam_sub = lam_til(sp,:);
        etil_sub = e_til(:,sp);

        lam1_sub = sum(repmat(lam_sub(:,1)',T,1).*etil_sub,2);   
        lam2_sub = sum(repmat(lam_sub(:,2)',T,1).*etil_sub,2);

        gamma_HAC_temp_no_order = [lam1_sub'*lam1_sub, lam1_sub'*lam2_sub
            lam2_sub'*lam1_sub, lam2_sub'*lam2_sub]/(n_sub*T);
        gamma_HAC_temp2_no_order = gamma_HAC_temp2_no_order + gamma_HAC_temp_no_order;

        sp_order = randsample(n-n_sub+1,1);
        lam_sub_order = lam_til(sp_order:sp_order+n_sub-1,:);
        etil_sub_order = e_til(:,sp_order:sp_order+n_sub-1);

        lam1_sub_order = sum(repmat(lam_sub_order(:,1)',T,1).*etil_sub_order,2);   
        lam2_sub_order = sum(repmat(lam_sub_order(:,2)',T,1).*etil_sub_order,2);

        gamma_HAC_temp_yes_order = [lam1_sub_order'*lam1_sub_order, lam1_sub_order'*lam2_sub_order
            lam2_sub_order'*lam1_sub_order, lam2_sub_order'*lam2_sub_order]/(n_sub*T);
        gamma_HAC_temp2_yes_order = gamma_HAC_temp2_yes_order + gamma_HAC_temp_yes_order;

    end;

    gamma_HAC_no_order = gamma_HAC_temp2_no_order/O;

    AV_f_HAC_no_order = (V_til\gamma_HAC_no_order)/V_til;    
    B2_HAC_no_order = z_hat(end,:)*AV_delta*z_hat(end,:)'/T + a_hat'*AV_f_HAC_no_order*a_hat/n;

    PI_cm_HAC_no_order = [yh_cm_hat(T,1) - 1.96*sqrt(B2_HAC_no_order), yh_cm_hat(T,1) + 1.96*sqrt(B2_HAC_no_order)];
    PI_HAC_no_order = [yh_cm_hat(T,1) - 1.96*sqrt(B2_HAC_no_order + varhat_eps ), yh_cm_hat(T,1) + 1.96*sqrt(B2_HAC_no_order + varhat_eps)];

    cnt_HAC_cm_no_order(c,1) = (yh_cm(T,1)>PI_cm_HAC_no_order(1,1))*(yh_cm(T,1)<PI_cm_HAC_no_order(1,2));
    cnt_HAC_no_order(c,1) = (yh(T,1)>PI_HAC_no_order(1,1))*(yh(T,1)<PI_HAC_no_order(1,2));

    gamma_HAC_yes_order = gamma_HAC_temp2_yes_order/O;

    AV_f_HAC_yes_order = (V_til\gamma_HAC_yes_order)/V_til;    
    B2_HAC_yes_order = z_hat(end,:)*AV_delta*z_hat(end,:)'/T + a_hat'*AV_f_HAC_yes_order*a_hat/n;

    PI_cm_HAC_yes_order = [yh_cm_hat(T,1) - 1.96*sqrt(B2_HAC_yes_order), yh_cm_hat(T,1) + 1.96*sqrt(B2_HAC_yes_order)];
    PI_HAC_yes_order = [yh_cm_hat(T,1) - 1.96*sqrt(B2_HAC_yes_order + varhat_eps ), yh_cm_hat(T,1) + 1.96*sqrt(B2_HAC_yes_order + varhat_eps)];

    cnt_HAC_cm_yes_order(c,1) = (yh_cm(T,1)>PI_cm_HAC_yes_order(1,1))*(yh_cm(T,1)<PI_cm_HAC_yes_order(1,2));
    cnt_HAC_yes_order(c,1) = (yh(T,1)>PI_HAC_yes_order(1,1))*(yh(T,1)<PI_HAC_yes_order(1,2));

end;

emp_cvg_cm_true = sum(cnt_cm_true)/C
emp_cvg_true = sum(cnt_true)/C

emp_cvg_cm = sum(cnt_cm)/C
emp_cvg = sum(cnt)/C
% % 
emp_cvg_con_cm = sum(cnt_con_cm)/C
emp_cvg_con = sum(cnt_con)/C

emp_cvg_HAC_cm_yes_order = sum(cnt_HAC_cm_yes_order)/C
emp_cvg_HAC_yes_order = sum(cnt_HAC_yes_order)/C

emp_cvg_HAC_cm_no_order = sum(cnt_HAC_cm_no_order)/C
emp_cvg_HAC_no_order = sum(cnt_HAC_no_order)/C
